﻿using System;
using System.IO;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Media.Media3D;
using System.Windows.Navigation;
using System.Windows.Shapes;
using System.ComponentModel;
using System.Threading;
using System.Windows.Threading;
using System.Runtime.InteropServices;
using System.Diagnostics;
using _3DTools;

#if REPRAP
using RepRap;
#endif

namespace Repic
{
	/// <summary>
	/// Interaction logic for MainWindow.xaml
	/// </summary>
	public partial class MainWindow : Window, IPrinter
	{
		const double INCHES_TO_MM = 2.54;

		const double EPSILON = 0.01;	// 1/10th resolution of the machine

		const int STEP_SIZE	= 16;
		const int STEP_RATE = 200;			// for a full circle
		const int NUM_PULLEY_TEETH = 8;
		const int NUM_GEAR_TEETH = 21;		// number of teeth on the z axis gear
		const double Z_AXIS_PITCH = 1.25; // mm
		const double X_STEPS_PER_MM = STEP_SIZE * STEP_RATE / 5.0 / NUM_PULLEY_TEETH;
		const double Y_STEPS_PER_MM = STEP_SIZE * STEP_RATE / 5.0 / NUM_PULLEY_TEETH;
		const double Z_STEPS_PER_MM = 6667; //STEP_SIZE * STEP_RATE * NUM_PULLEY_TEETH / NUM_GEAR_TEETH / Z_AXIS_PITCH; (?)
		const double E_STEPS_PER_MM = 27.0;

		private int _XStepper = 0;
		private int XStepper
		{
			get {return this._XStepper;}
			set
			{
				if (this._XStepper != value)
				{
					this._XStepper = value;
					SendUsbPacket(new XStepperPacket(this._XStepper));
				}
			}
		}

		private int _YStepper = 0;
		private int YStepper
		{
			get {return this._YStepper;}
			set
			{
				if (this._YStepper != value)
				{
					this._YStepper = value;
					SendUsbPacket(new YStepperPacket(this._YStepper));
				}
			}
		}

		private int _ZStepper = 0;
		private int ZStepper
		{
			get {return this._ZStepper;}
			set
			{
				if (this._ZStepper != value)
				{
					this._ZStepper = value;
					SendUsbPacket(new ZStepperPacket(this._ZStepper));
				}
			}
		}

		private int _EStepper = 0;
		private int EStepper
		{
			get {return this._EStepper;}
			set
			{
				if (this._EStepper != value)
				{
					this._EStepper = value;
					SendUsbPacket(new EStepperPacket(this._EStepper));
				}
			}
		}

		private bool _ExtruderHeater = false;
		private bool ExtruderHeater
		{
			get {return this._ExtruderHeater;}
			set
			{
				if (this._ExtruderHeater != value)
				{
					this._ExtruderHeater = value;
					SendUsbPacket(new ExtruderHeaterPacket(this._ExtruderHeater));
				}
			}
		}

		private bool _XEndStop = false;
		private bool XEndStop
		{
			get { return this._XEndStop; }
			set { this.XEndStopControl.Fill = value ? Brushes.Yellow : Brushes.DarkGray; }
		}

		private bool _YEndStop = false;
		private bool YEndStop
		{
			get { return this._YEndStop; }
			set { this.YEndStopControl.Fill = value ? Brushes.Yellow : Brushes.DarkGray; }
		}

		private bool _ZEndStop = false;
		private bool ZEndStop
		{
			get { return this._ZEndStop; }
			set { this.ZEndStopControl.Fill = value ? Brushes.Yellow : Brushes.DarkGray; }
		}

		private int _ExtruderADC = 0xffff;
		private int ExtruderADC
		{
			get { return this._ExtruderADC; }
			set {
				this._ExtruderADC = value;
				
				// convert to a temperature and set the thermometer control
				double v = value * vadc / 1023;						// convert the 10 bit ADC value to a voltage
				double r = rs * v / (vs - v);						// resistance of thermistor
				double temp = (beta / Math.Log(r / k)) - 273.15;	// temperature
				this.ExtruderThermometer.CurrentTemp = (int)temp;
				labelExtruderThermistor.Content = String.Format("{0} - {1:0}", value, temp);
			}
		}

		private int SlotsAvailable = 0;
		PrintSequence Sequence = null;
		Queue<ExtrudeData> Extrudes = new Queue<ExtrudeData>();

		bool _Printing = false;
		bool Printing
		{
			get
			{
				return this._Printing;
			}
			set
			{
				this._Printing = value;
				if (value)
					this.btnPrint.Content = "Stop";
				else
					this.btnPrint.Content = "Print";
			}
		}

		// worker thread for background processing
		BackgroundWorker WorkerThread;

		List<GeometryModel3D> CurrentChildModels = new List<GeometryModel3D>();

		DiffuseMaterial MeshMaterial = new DiffuseMaterial(Brushes.CornflowerBlue);
		//material.Brush = Brushes.OrangeRed;
		//material.Brush = Brushes.Blue;
		//material.Brush = Brushes.White;

		RepicMode _CurrentMode = RepicMode.Unknown;
		RepicMode CurrentMode
		{
			get { return this._CurrentMode; }
			set
			{
				if (value != this._CurrentMode)
				{
					this._CurrentMode = value;
					this.menuVirtualMendel.IsChecked = (value == RepicMode.VirtualMendel);
					this.menuHardwareTest.IsChecked = (value == RepicMode.HardwareTest);
					this.menuPrint.IsChecked = (value == RepicMode.Print);
					if (value == RepicMode.HardwareTest)
					{
						this.ModelTrackball.IsEnabled = false;
						this.ModelTrackball.Visibility = System.Windows.Visibility.Collapsed;
						this.HardwareTestViewbox.IsEnabled = true;
						this.HardwareTestViewbox.Visibility = System.Windows.Visibility.Visible;
					}
					else
					{
						this.ModelTrackball.IsEnabled = true;
						this.ModelTrackball.Visibility = System.Windows.Visibility.Visible;
						this.HardwareTestViewbox.IsEnabled = false;
						this.HardwareTestViewbox.Visibility = System.Windows.Visibility.Collapsed;
					}
				}
			}
		}

		GeometryModel3D CurrentModel;
		Usb Usb;

		public MainWindow()
		{			
			InitializeComponent();
			this.DataContext = this;
		}

		private void Window_Loaded(object sender, RoutedEventArgs e)
		{
			// load the gcodes
			//this.Sequence = new PrintSequence(this, @"c:\temp\prusa\pla-bushing.gcode");
			ResetModel();

			this.CurrentMode = RepicMode.VirtualMendel;

			// initialize the usb interface
			this.Usb = new Usb(this);
			this.Usb.OnStatusChanged += this.OnUsbStatusChanged;

			// initialize the worker thread
			this.WorkerThread = new BackgroundWorker();
			this.WorkerThread.DoWork += new DoWorkEventHandler(WorkerThread_DoWork);
			this.WorkerThread.RunWorkerAsync();
		}

		void WorkerThread_DoWork(object sender, DoWorkEventArgs e)
		{
			EventWaitHandle wait = new AutoResetEvent(false);

			while (true)
			{
				// rest a bit
				Thread.Sleep(1000 / 60);

				GetRepRapStatus();

				// nothing to do if we're not printing
				if (!this.Printing)
					continue;

				// if we're at the end of the print sequence then stop printing
				if ((this.Sequence != null) && this.Sequence.EndOfSequence)
				{
					wait.Reset();
					this.Dispatcher.BeginInvoke(DispatcherPriority.Normal, (ThreadStart)delegate()
					{
						this.Printing = false;
						wait.Set();
					});
					wait.WaitOne();
					continue;
				}

				// keep printing for 1/60th of a second  (or until the queue fills up if we're printing to hardware)
				DateTime start = DateTime.Now;
				do
				{
					if ((this.CurrentMode== RepicMode.Print) && (this.SlotsAvailable <= 0))
						break;
					this.Sequence.Step();
				} while (((DateTime.Now - start) < TimeSpan.FromMilliseconds(1000 / 60)) && (!this.Sequence.EndOfSequence));

				// draw any extrusions we have				
				wait.Reset();
				this.Dispatcher.BeginInvoke(DispatcherPriority.Normal, (ThreadStart)delegate()
				{
					// draw any extrusions
					while (this.Extrudes.Count > 0)
						AddExtrusion(this.Extrudes.Dequeue());
					wait.Set();
				});
				wait.WaitOne();				
			}
		}

		void UpdatePrint()
		{
			// if we're at the end of the print sequence then stop printing
			if (this.Sequence.EndOfSequence)
			{
				this.Dispatcher.BeginInvoke(DispatcherPriority.Normal, (ThreadStart)delegate()
				{
					this.Printing = false;
				});
				return;
			}

			// make sure there are slots available
			if ((this.CurrentMode != RepicMode.Print) || (this.SlotsAvailable > 0))
			{
				// print a bit more
				lock (this.Extrudes)
				{
					// keep printing for 1/60th of a second
					DateTime start = DateTime.Now;
					do
					{
						this.Sequence.Step();
					} while (((DateTime.Now - start) < TimeSpan.FromMilliseconds(1000 / 60)) &&
								(!this.Sequence.EndOfSequence) &&
								(this.SlotsAvailable > 0));
				}

				// draw any extrusions we have
				EventWaitHandle wait = new AutoResetEvent(false);
				this.Dispatcher.BeginInvoke(DispatcherPriority.Normal, (ThreadStart)delegate()
				{
					// draw any extrusions
					while (this.Extrudes.Count > 0)
						AddExtrusion(this.Extrudes.Dequeue());
					wait.Set();
				});
				wait.WaitOne();
			}
		}


		private void OnUsbStatusChanged(Usb sender, ConnectionStatus oldStatus, ConnectionStatus newStatus)
		{
			this.Dispatcher.Invoke(DispatcherPriority.Normal, new Action(delegate()
			{
				switch (newStatus)
				{
					case ConnectionStatus.Unknown:
						this.Status.Text = "Status: Unknown";
						break;

					case ConnectionStatus.Disconnected:
						this.Status.Text = "Status: Disconnected";
						break;

					case ConnectionStatus.Broken:
						this.Status.Text = "Status: Broken";
						break;

					case ConnectionStatus.Connected:
						this.Status.Text = "Status: Connected";
						break;
				}
			}
			));
		}

		private void ResetModel()
		{
			Model3DGroup group = this.Resources["MendelGroup"] as Model3DGroup;
			foreach (GeometryModel3D child in this.CurrentChildModels)
				group.Children.Remove(child);
			this.CurrentChildModels.Clear();
			CreateNewModel();
		}

		private void CreateNewModel()
		{
			this.CurrentModel = new GeometryModel3D();
			this.CurrentModel.Material = this.MeshMaterial;
			MeshGeometry3D mesh = new MeshGeometry3D();
			this.CurrentModel.Geometry = mesh;
			mesh.Positions = new Point3DCollection();
			mesh.Normals = new Vector3DCollection();
			mesh.TriangleIndices = new Int32Collection();

			Model3DGroup group = this.Resources["MendelGroup"] as Model3DGroup;
			group.Children.Add(this.CurrentModel);
			this.CurrentChildModels.Add(this.CurrentModel);
		}
		
		int extrusionsPerTick = 100;

		private void OnLoad(object sender, RoutedEventArgs e)
		{			
			this.CurrentFile.Text = "";

			// Create OpenFileDialog
			Microsoft.Win32.OpenFileDialog dlg = new Microsoft.Win32.OpenFileDialog();          
 
			// Set filter for file extension and default file extension
			dlg.DefaultExt = ".txt";
			dlg.Filter = "GCode file documents (.gcode)|*.gcode";
 
			// Display OpenFileDialog by calling ShowDialog method
			Nullable<bool> result = dlg.ShowDialog();
 
			// Get the selected file name and display in a TextBox
			if (result != true)
				return;

			using (new CursorHelper(this, Cursors.Wait))
			{
				LoadFile(dlg.FileName);
				this.CurrentFile.Text = dlg.FileName;
			}
		}

		private void CommandBinding_Executed(object sender, ExecutedRoutedEventArgs e)
		{
			this.Close();
		}

		private void OnClosing(object sender, System.ComponentModel.CancelEventArgs e)
		{
		}

		private void AddExtrusion(ExtrudeData e)
		{
			if ((e.Start.X == e.End.X) && (e.Start.Y == e.End.Y) && (e.Start.Z == e.End.Z))
				return;
			if (e.Start.E >= e.End.E)
				return;

			float s = 1 / 64.0f;		// geometry render scale
			double dx = e.End.X - e.Start.X;
			double dy = e.End.Y - e.Start.Y;
			double dz = e.End.Z - e.Start.Z;
			double de = e.End.E - e.Start.E;
			double distance = Math.Sqrt(dx * dx + dy * dy + dz * dz);
			double width = 0.005 * de / distance;

			Point3D p1 = new Point3D(17 + e.Start.X * s, e.Start.Y * s, e.Start.Z * s + 3.8f);
			Point3D p2 = new Point3D(17 + e.End.X * s, e.End.Y * s, e.End.Z * s + 3.8f);
			Vector3D up = new Vector3D(0, 0, 1);
			Vector3D y = p2 - p1;
			y.Normalize();
			Vector3D x = Vector3D.CrossProduct(y, up);
			x.Normalize();
			Vector3D z = Vector3D.CrossProduct(x, y);
			z.Normalize();
			Point3D a1 = p1 + width * z;
			Point3D a2 = p1 + width * x;
			Point3D a3 = p1 - width * x;
			Point3D b1 = p2 + width * z;
			Point3D b2 = p2 + width * x;
			Point3D b3 = p2 - width * x;
			AddPrim(a1, a2, b1, b2, z, x, z, x);
			AddPrim(a3, a1, b3, b1, -x, z, -x, z);
		}

		private void AddPrim(Point3D p0, Point3D p1, Point3D p2, Point3D p3,
							 Vector3D n0, Vector3D n1, Vector3D n2, Vector3D n3)
		{
			MeshGeometry3D mesh = this.CurrentModel.Geometry as MeshGeometry3D;						
			int index = mesh.Positions.Count;

			mesh.Positions.Add(p0);
			mesh.Positions.Add(p1);
			mesh.Positions.Add(p2);
			mesh.Positions.Add(p3);
			mesh.Normals.Add(n0);
			mesh.Normals.Add(n1);
			mesh.Normals.Add(n2);
			mesh.Normals.Add(n3);
			mesh.TriangleIndices.Add(index + 0);
			mesh.TriangleIndices.Add(index + 1);
			mesh.TriangleIndices.Add(index + 2);
			mesh.TriangleIndices.Add(index + 1);
			mesh.TriangleIndices.Add(index + 3);
			mesh.TriangleIndices.Add(index + 2);

			// don't let the mesh get too large, otherwise we'll be spending all our time rebuilding geometry.
			if (mesh.Positions.Count > 1000)
				CreateNewModel();
		}
		
		int frameNum = 0;

		private void SaveFrame()
		{
			// save to bitmaps
			RenderTargetBitmap bitmap = new RenderTargetBitmap(
				(int)this.mainViewport.ActualWidth, (int)this.mainViewport.ActualHeight,
				96, 96, PixelFormats.Pbgra32);
			bitmap.Render(this.mainViewport);
			PngBitmapEncoder image = new PngBitmapEncoder();
			image.Frames.Add(BitmapFrame.Create(bitmap));
			string filename = String.Format(@"Frames\frame_{0}.png", frameNum++);
			using (Stream fs = File.Create(filename))
				image.Save(fs);
		}

		private void LoadFile(string filename)
		{
			// load the file
			try
			{				
				this.Sequence = new PrintSequence(this, filename);
			}
			catch (Exception ex)
			{
				MessageBox.Show("Error loading gcode file: " + ex.Message);
				this.Sequence = null;
				return;
			}

			// get the model ready
			ResetModel();
		}

		static double r0 = 100000;									// stated resistance, e.g. 10K
		static double r1 = 4700;
		static double r2 = 10000;
		static double t0 = 25.0 + 273.15;							// temperature at stated resistance, e.g. 25C
		static double beta = 4066;									// stated beta, e.g. 3500
		static double vadc = 5.0;									// ADC reference
		static double vcc = 5.0;									// supply voltage to potential divider
		static double vs = r1 * vcc / (r1 + r2);					// effective bias voltage
		static double rs = r1 * r2 / (r1 + r2);					// effective bias impedance
		static double k = r0 * Math.Exp(-beta / t0);				// constant part of calculation

		static ushort TempToADC(float t)
		{
			double r = r0 * Math.Exp(beta * (1 / (t + 273.15) - 1 / t0)); // resistance of the thermistor
			double v = vs * r / (rs + r);	// the voltage at the potential divider
			return (ushort)Math.Round(v / vadc * 1023);  // the ADC reading
		}

		static int ADCToTemp(int adc)
		{
			double v = adc * vadc / 1023;						// convert the 10 bit ADC value to a voltage
			double r = rs * v / (vs - v);						// resistance of thermistor
			double temp = (beta / Math.Log(r / k)) - 273.15;	// temperature
			return (int)Math.Round(temp);
		}

		private void SendUsbPacket(RepRapPacket msg)
		{
			if (this.Usb == null)
				return;
			if (this.Usb.Status != ConnectionStatus.Connected)
				return;
			try
			{
				this.Usb.SendPacket(msg);
			}
			catch (Exception)
			{
			}
		}

		private void XAxisLeftPressedDown(object sender, MouseButtonEventArgs e) { this.XStepper = -1; }
		private void XAxisLeftPressedUp(object sender, MouseButtonEventArgs e) { this.XStepper = 0; }
		private void XAxisRightPressedDown(object sender, MouseButtonEventArgs e) { this.XStepper = 1; }
		private void XAxisRightPressedUp(object sender, MouseButtonEventArgs e) { this.XStepper = 0; }

		private void YAxisBackPressedDown(object sender, MouseButtonEventArgs e) { this.YStepper = -1; }
		private void YAxisBackPressedUp(object sender, MouseButtonEventArgs e) { this.YStepper = 0; }
		private void YAxisForwardPressedDown(object sender, MouseButtonEventArgs e) { this.YStepper = 1; }
		private void YAxisForwardPressedUp(object sender, MouseButtonEventArgs e) { this.YStepper = 0; }

		private void ZAxisDownPressedDown(object sender, MouseButtonEventArgs e) { this.ZStepper = -1; }
		private void ZAxisDownPressedUp(object sender, MouseButtonEventArgs e) { this.ZStepper = 0; }
		private void ZAxisUpPressedDown(object sender, MouseButtonEventArgs e) { this.ZStepper = 1; }
		private void ZAxisUpPressedUp(object sender, MouseButtonEventArgs e) { this.ZStepper = 0; }

		private void EAxisReversePressedDown(object sender, MouseButtonEventArgs e) { this.EStepper = -1; }
		private void EAxisReversePressedUp(object sender, MouseButtonEventArgs e) { this.EStepper = 0; }
		private void EAxisForwardPressedDown(object sender, MouseButtonEventArgs e) { this.EStepper = 1; }
		private void EAxisForwardPressedUp(object sender, MouseButtonEventArgs e) { this.EStepper = 0; }

		private void OnAbout(object sender, RoutedEventArgs e)
		{
			new AboutBox().ShowDialog();
		}

		private void OnVirtualMendel(object sender, RoutedEventArgs e)
		{
			this.CurrentMode = RepicMode.VirtualMendel;
		}

		private void OnHardwareTest(object sender, RoutedEventArgs e)
		{
			this.CurrentMode = RepicMode.HardwareTest;
		}

		private void OnPrint(object sender, RoutedEventArgs e)
		{
			this.CurrentMode = RepicMode.Print;
		}

		private void OnExtruderHeaterClicked(object sender, RoutedEventArgs e)
		{
			this.ExtruderHeater = !this.ExtruderHeater;
		}

		private void OnDeviceSettings(object sender, RoutedEventArgs e)
		{
			// if we're not connected then we can't do anything
			if (this.Usb.Status != ConnectionStatus.Connected)
			{
			    MessageBox.Show("No RepRap device connected.");
			    return;
			}

			try
			{
				// show the dialog box
				SettingsDlg dlg = new SettingsDlg();
				dlg.ExtruderTemp = GetRepRapTargetTemp();
				dlg.Owner = this;
				if (dlg.ShowDialog().Value)
					return;
				SetRepRapTargetTemp(dlg.ExtruderTemp);
			}
			catch
			{
				if (this.Usb.Status != ConnectionStatus.Connected)
					MessageBox.Show("RepRap device no longer connected, changes discarded.");
				else
					MessageBox.Show("Error occurred sending settings to RepRap device, changes discarded.");
			}
		}

		private void OnPrintCurrentFile(object sender, RoutedEventArgs e)
		{
			if (this.Sequence == null)
				MessageBox.Show("Nothing to print!");
			else if ((this.Sequence.EndOfSequence) && !this.Printing)
			{
				ResetModel();
				this.Sequence.Reset();
				this.Printing = true;
			}
			else
				this.Printing = !this.Printing;
		}

		protected void GetRepRapStatus()
		{
			// tell the RepRap to send its status
			SendUsbPacket(new RequestStatusPacket());

			// wait for it to respond (TO DO: this needs to be made a bit more stable)
			RepRapPacket status = this.Usb.WaitForPacket();

			// process the data
			if (status.Data[1] == (byte)MessageID.MSGID_STATUS)
			{
				this.Dispatcher.BeginInvoke(DispatcherPriority.Normal, (ThreadStart)delegate()
				{
					this.XEndStop = (status.Data[2] != 0);
					this.YEndStop = (status.Data[3] != 0);
					this.ZEndStop = (status.Data[4] != 0);
					this.ExtruderADC = (ushort)((int)status.Data[5] + (int)status.Data[6] * 256);
					this.SlotsAvailable = status.Data[7];
				});
			}
		}

		protected void SetRepRapTargetTemp(int temp)
		{
			SendUsbPacket(new SetTargetTempPacket(temp, TempToADC(temp)));
		}

		protected int GetRepRapTargetTemp()
		{
			// tell the RepRap to send its status
			SendUsbPacket(new RequestTargetTempPacket());

			// wait for it to respond (TO DO: this needs to be made a bit more stable)
			RepRapPacket packet = this.Usb.WaitForPacket();

			// process the data
			if (packet.Data[1] == (byte)MessageID.MSGID_TEMP_SETTING)
				return packet.Data[2];

			return 220;	// default
		}

		/*
		 * Print callbacks
		 */
		public void GoHome()
		{
			// if we're printing then send it to the reprap
			if (this.CurrentMode == RepicMode.Print)
			{
				SendUsbPacket(new GoHomePacket());
				this.SlotsAvailable--;
			}
		}

		public void ControlledMove(ExtrudeData extrudeData)
		{
			// if we're printing then send it to the reprap
			if (this.CurrentMode == RepicMode.Print)
			{
				SendExtrusion(extrudeData);
				this.SlotsAvailable--;
			}
			
			// add it to the queue
			this.Extrudes.Enqueue(extrudeData);
		}

		public void Dwell(double dwellTime)
		{
			// if we're printing then send it to the reprap
			if (this.CurrentMode == RepicMode.Print)
			{
				SendUsbPacket(new DwellPacket(dwellTime));
				this.SlotsAvailable--;
			}
		}

		// based on cartesian_dda::set_target in the official firmware
		void SendExtrusion(ExtrudeData extrudeData)
		{
			// calculate our units, adjusting for imperial scale if we need to
			double scale = (this.Sequence.CurrentUnitType == UnitType.Imperial) ? INCHES_TO_MM : 1.0;
			Vector5D units = new Vector5D(
				X_STEPS_PER_MM * scale, Y_STEPS_PER_MM * scale, Z_STEPS_PER_MM * scale, E_STEPS_PER_MM * scale, 1.0);

			// calculate distance to move along each axis
			Vector5D delta = extrudeData.End - extrudeData.Start;

			// calculate how far we're going to move the extrude head
			Vector5D squares = delta * delta;
			double distance = squares.X + squares.Y + squares.Z;

			// if it's 0 then the only thing changing is e
			if (distance < EPSILON * EPSILON)
				distance = squares.E;

			// if it's still zero then the only thing changing is the feed rate
			if (distance < EPSILON * EPSILON)
			{
				//distance = squares.F;
				return;
			}

			// calculate the total distance
			distance = Math.Sqrt(distance);

			// calculate the time (in interrupt ticks) based on the feedrate
			if (extrudeData.End.F <= 0)
				return;
			int totalTicks = (int)(60 * 10000 * distance / extrudeData.End.F);
			if (totalTicks <= 0)
				return;

			// calculate our current, target and delta steps
			Vector5D currentSteps = ToSteps(units, extrudeData.Start);
			Vector5D targetSteps = ToSteps(units, extrudeData.End);
			//Vector5D deltaSteps = Vector5D.Abs(targetSteps - currentSteps);
			Vector5D deltaSteps = targetSteps - currentSteps;
			short xTicks = (short)deltaSteps.X;
			short yTicks = (short)deltaSteps.Y;
			short zTicks = (short )deltaSteps.Z;
			short eTicks = (short )deltaSteps.E;
			if ((xTicks == 0) && (yTicks == 0) && (zTicks == 0) && (eTicks == 0))
				return;
			if (eTicks < 0)
				eTicks = 0;

			// send the data to the repap
			SendUsbPacket(new ControlledMovePacket(totalTicks, xTicks, yTicks, zTicks, eTicks));
		}

		Vector5D ToSteps(Vector5D units, Vector5D v)
		{
			v *= units;
			v.X = (int)v.X;
			v.Y = (int)v.Y;
			v.Z = (int)v.Z;
			v.E = (int)v.E;
			v.F = (int)v.F;
			return v;
		}

	}

	public enum RepicMode
	{
		Unknown,
		VirtualMendel,
		HardwareTest,
		Print
	}
	
}
