﻿using System;
using System.IO;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Diagnostics;

namespace Repic
{
	public class PrintSequence
	{
		public IPrinter Printer;
		public string[] GCodes = null;
		public int CurrentInstruction = 0;
		public UnitType CurrentUnitType = UnitType.Metric;
		public PositioningType CurrentPositioningType = PositioningType.Absolute;
		public Vector5D CurrentPosition = new Vector5D(0.0, 0.0, 0.0, 0.0, 0.0);
		public Vector5D NextPosition = new Vector5D(0.0, 0.0, 0.0, 0.0, 0.0);

		public bool EndOfSequence { get { return (this.CurrentInstruction >= this.GCodes.Length); } }

		private delegate void PrintCommandHandler(string args);
		private Dictionary<int, PrintCommandHandler> GCodeHandlers = new Dictionary<int, PrintCommandHandler>();
		private Dictionary<int, PrintCommandHandler> MCodeHandlers = new Dictionary<int, PrintCommandHandler>();

		public PrintSequence(IPrinter printer, string filename)
		{
			this.Printer = printer;
			this.GCodes = File.ReadAllLines(filename);
			SetUpCommandHandlers();
			Reset();
		}

		public void Reset()
		{
			this.CurrentInstruction = 0;
			this.CurrentUnitType = UnitType.Metric;
			this.CurrentPositioningType = PositioningType.Absolute;
			this.CurrentPosition = new Vector5D(0.0, 0.0, 0.0, 0.0, 0.0);
			this.NextPosition = new Vector5D(0.0, 0.0, 0.0, 0.0, 0.0);
		}

		public void SetUpCommandHandlers()
		{
			// gcodes
			GCodeHandlers[0] = RapidMove;
			GCodeHandlers[1] = ControlledMove;
			GCodeHandlers[28] = GoHome;
			GCodeHandlers[4] = Dwell;
			GCodeHandlers[20] = SetImperial;
			GCodeHandlers[21] = SetMetric;
			GCodeHandlers[90] = AbsolutePositioning;
			GCodeHandlers[91] = IncrementalPositioning;
			GCodeHandlers[92] = SetPosition;

			// mcodes
			MCodeHandlers[0] = ShutDown;
			MCodeHandlers[140] = SetBedTemp;
			MCodeHandlers[113] = SetExtruderPot;
			MCodeHandlers[109] = SetExtruderTempWait;
			MCodeHandlers[107] = CoolerOff;
			MCodeHandlers[106] = CoolerOn;
			MCodeHandlers[104] = SetExtruderTemp;
		}

		public void Step()
		{
			Match match;

			// if we've finished printing then don't bother
			if (this.EndOfSequence)
				return;

			// get the next instruction
			string instr = this.GCodes[this.CurrentInstruction++];

			// check for a g-code
			match = Regex.Match(instr, @"^G(?<code>\d+)\b(?<args>.*)");
			if (match.Groups["code"].ToString() != "")
			{
				int code = Convert.ToInt32(match.Groups["code"].ToString());
				Debug.Assert(GCodeHandlers.ContainsKey(code), "Unknown G-Code: " + code.ToString());
				if (GCodeHandlers.ContainsKey(code))
				{
					string args = match.Groups["args"].ToString();
					GCodeHandlers[code](args);
				}
			}

			// check for an m-code
			match = Regex.Match(instr, @"^M(?<code>\d+)\b(?<args>.*)");
			if (match.Groups["code"].ToString() != "")
			{
				int code = Convert.ToInt32(match.Groups["code"].ToString());
				Debug.Assert(MCodeHandlers.ContainsKey(code), "Unknown M-Code: " + code.ToString());
				if (MCodeHandlers.ContainsKey(code))
				{
					string args = match.Groups["args"].ToString();
					MCodeHandlers[code](args);
				}
			}
		}

		private void ParseNumberField(string args, string field, ref double val)
		{
			Match match = Regex.Match(args, field + @"(?<val>[0-9+-.]+\b)");
			if (match.Groups["val"].ToString() != "")
				val = Convert.ToDouble(match.Groups["val"].ToString());
		}

		/*
		 * Command Handlers
		 */

		private void RapidMove(string args)
		{
			// just do a regular controlled move for now
			ControlledMove(args);
		}

		private void ControlledMove(string args)
		{
			this.NextPosition = this.CurrentPosition;
			ParseNumberField(args, "X", ref this.NextPosition.X);
			ParseNumberField(args, "Y", ref this.NextPosition.Y);
			ParseNumberField(args, "Z", ref this.NextPosition.Z);
			ParseNumberField(args, "E", ref this.NextPosition.E);
			ParseNumberField(args, "F", ref this.NextPosition.F);
			if (this.Printer != null)
				this.Printer.ControlledMove(new ExtrudeData(this.CurrentPosition, this.NextPosition));
			this.CurrentPosition = this.NextPosition;
		}

		private void GoHome(string args)
		{
			if (this.Printer != null)
				this.Printer.GoHome();
		}

		private void Dwell(string args)
		{
			double dwellTime = 0;
			ParseNumberField(args, "P", ref dwellTime);
			if (this.Printer != null)
				this.Printer.Dwell(dwellTime);
		}

		private void SetImperial(string args)
		{
			this.CurrentUnitType = UnitType.Imperial;
		}

		private void SetMetric(string args)
		{
			this.CurrentUnitType = UnitType.Metric;
		}

		private void AbsolutePositioning(string args)
		{
			this.CurrentPositioningType = PositioningType.Absolute;
		}

		private void IncrementalPositioning(string args)
		{
			this.CurrentPositioningType = PositioningType.Incremental;
		}

		private void SetPosition(string args)
		{
			ParseNumberField(args, "X", ref this.CurrentPosition.X);
			ParseNumberField(args, "Y", ref this.CurrentPosition.Y);
			ParseNumberField(args, "Z", ref this.CurrentPosition.Z);
			ParseNumberField(args, "E", ref this.CurrentPosition.E);
		}

		private void ShutDown(string args)
		{
			// ignored
		}

		private void SetBedTemp(string args)
		{
			// ignored
		}

		private void SetExtruderPot(string args)
		{
			// ignored
		}

		private void SetExtruderTemp(string args)
		{
			// ignored
		}

		private void SetExtruderTempWait(string args)
		{
			// ignored
		}

		private void CoolerOff(string args)
		{
			// ignored
		}

		private void CoolerOn(string args)
		{
			// ignored
		}
	}

	public enum UnitType
	{
		Imperial,
		Metric
	}

	public enum PositioningType
	{
		Absolute,
		Incremental
	}
}
